/*
	
	Basic image/bitmap operations
		by Grzegorz Tanski

*/

#include <iostream>

#include "../kraysdk/kray.h"
#include "../kraysdk/symbol/bitmap.h"
#include "../kraysdk/type/vector.h"
#include "../kraysdk/buffer/binary.h"
#include "../kraysdk/proto/direct.h"
#include "../kraysdk/commands.h"

#include "../SharedSources/libraryDir.h"

using namespace Kray;

int kraySdkExampleMain(int argc, const char** argv){	// main function is in startup.cpp
	Library klib(KRAY_DIRECTORY);
	Instance* kinst=klib.createInstance();
	
	if (kinst){
		CommandBinaryBuffer tbuf(kinst);
		DirectPrototyper prot(tbuf);

		{	// create Kray bitmap and fill it with content
			const int width=300;
			const int height=400;

			BitmapSymbol bmps(prot,width,height,PF_CHAR_RGBA);

			Bitmap* bmp=bmps.getDirect();
		
			float cx=0;
			float cy=-0.5;
			float scale=0.01F;

			const int steps=16;
			const int bailOut=100;

			const bool julia=false;
			const float jim=-0.1F;
			const float jre=0.65F;
	
			for (int y=0 ; y<height ; y++){
				float cre=(y-height/2)*scale+cy;
				for (int x=0 ; x<width ; x++){
					float cim=(x-width/2)*scale+cx;

					float vre=cre;
					float vim=cim;

					int step=0;
					for (; step<steps ; step++){

						float t=vre*vre-vim*vim;
						vim=2*vre*vim;
						vre=t;	
		
						if (julia){
							vre+=jre;
							vim+=jim;
						}else{
							vre+=cre;
							vim+=cim;
						}

						if ((vre*vre+vim*vim)>bailOut){
							break;
						}
					}

					bmp->setPixel(x,y,Vector((float)step/steps,(float)step/steps*3/2,(float)step/steps/2,1));
				}
			}
		
			prot.view_select("Newly created bitmap");
			prot.bitmapOp_show(bmps);
		}

		{
			const int width=400;
			const int height=300;

			unsigned char* pixels=new unsigned char[4*width*height];	// RGBA , 4 bytes per pixel

			unsigned char* pixPtr=pixels;

			for (int y=0 ; y<height ; y++){
				for (int x=0 ; x<width ; x++){
					*pixPtr++=(unsigned char)255;				// R
					*pixPtr++=(unsigned char)(x*256/width);		// G
					*pixPtr++=(unsigned char)0;					// B
					*pixPtr++=(unsigned char)(y*256/height);	// A
				}
			}

			// when we have bitmap data, we can create a copy of it in Kray
			// create a copy of existing bitmap data

			Bitmap bmp((void*)pixels,PF_CHAR_RGBA,width);	// temporary object that holds only reference to our bitmap, last parameter is length of bitmap line in pixels (can be higher the bitmap width in general)

			BitmapSymbol bmps(prot,bmp,width,height);

			delete[] pixels;	// we have a copy in Kray, now we can delete pixels

			prot.view_select("Copy of bitmap from existing pixels");
			prot.bitmapOp_showAlpha(bmps);	// shows semi transparent window with alpha (on platforms that supports it)
		}

		{
			BitmapSymbol bmps(PROJECT_DIRECTORY "Bitmaps/kray-logo.png",prot);

			if (prot.buffer().lastError()){	// file not found
				prot.buffer().clearError();		
			}else{
				prot.view_select("Bitmap from file");
				prot.bitmapOp_show(bmps);
			}
		}

		{
			extern void setupSimpleMeshScene(Kray::Prototyper& prot);

			setupSimpleMeshScene(prot);

			prot.view_select("");
			prot.sett_previewSize(Kray::Vector(0,0));
			prot.sett_buffersMask(1<<KRAY_NEEDBUFFER_BIT_ALPH|1<<KRAY_NEEDBUFFER_BIT_RGB);		// lets render alpha channel
			prot.render();

			BitmapSymbol render(prot,0,0,PF_CHAR_RGBA);
			prot.bitmapOp_copyView_rgba(render);		// this way we copy output without changing pixel format (PF_CHAR_RGBA)

			// we have same access to rendered image pixels like for other bitmaps
			KrayBitmapInfo bitmapInfo=render.getInfo();

			unsigned char* pixPtr=(unsigned char*)bitmapInfo.pixels;
			
			for (int y=0 ; y<bitmapInfo.sizy ; y++){
				for (int x=0 ; x<bitmapInfo.sizx ; x++){
					pixPtr+=3;									// skip RGBA
					*pixPtr=*pixPtr/2+127;						// A

					pixPtr++;
				}
			}
	
			prot.view_select("Bitmap from render output");
			prot.bitmapOp_showAlpha(render);
		}

		prot.view_select("");
		prot.pause(0);
	}else{
		std::cout << "Failed to load Kray instance" << std::endl;
	}

	return 0;
}

