#pragma once

#include "resource.h"

#include "../kraysdk/kray.h"
#include "../kraysdk/base/listener.h"
#include "../kraysdk/base/getinfo.h"
#include "../kraysdk/proto/direct.h"
#include "../kraysdk/proto/proto.h"
#include "../kraysdk/buffer/cmd.h"

// kray listener class
// InstanceLinkedListener has empty implementations of all methods and allows us to filter Kray log messages (detect render finish for example)
// we could use BounceBackListner as base class, so all unhandled events will return to Kray (and will launch their default implementation)

class ExampleKrayEventListener : public Kray::InstanceLinkedListener{
	HWND hWnd;		// set in WM_CREATE handle
	HBITMAP bitmap;	// holds render preview
	CRITICAL_SECTION criticalSection;	// bitmap is accessed by render thread and main thread, we need synchronization
	
	int previewSizeX;
	int previewSizeY;
	
	char isInterrupted;
	char windowCreated;

public:
	ExampleKrayEventListener(HWND hWnd=0):hWnd(hWnd){
		isInterrupted=0;
		bitmap=0;
		
		InitializeCriticalSection(&criticalSection);
	}
	~ExampleKrayEventListener(){
		freeBitmap();		
	}

	void* createWindow(){
		if (windowCreated) return 0;	// we support only one window

		return this;
	}
	void destroyWindow(void* windowHandle){
		if (windowHandle==this) windowCreated=0;
	}

	void setWindowHandle(HWND hWnd){
		this->hWnd=hWnd;
	}
	void freeBitmap(){
		if (bitmap){
			DeleteObject(bitmap);
			bitmap=0;
		}
	}	
	void interruptRender(){
		if (isInterrupted){
			PostMessage(hWnd,WM_DESTROY,0,0);
		}

		isInterrupted=1;
	}
	void repaint(PAINTSTRUCT* ps){	// called from WndProc WM_PAINT:
		if (bitmap){
			EnterCriticalSection(&criticalSection);

			HDC hdc=GetDC(hWnd);	
			HDC hdcMem=CreateCompatibleDC(hdc);
			SelectObject(hdcMem,bitmap);

			BitBlt(hdc,0,0,previewSizeX,previewSizeY,hdcMem,0,0,SRCCOPY);		
	
			ReleaseDC(hWnd,hdc);
			DeleteDC(hdcMem);
			
			LeaveCriticalSection(&criticalSection);
		}
	}
	void renderFinished(){	// kray calls this function when rendering process end (render finished or interupted);
		MessageBox(hWnd,TEXT("Render done"),TEXT("Info"),MB_ICONINFORMATION|MB_SETFOREGROUND);
		
		PostMessage(hWnd,WM_DESTROY,0,0);
	}
	void renderInterupted(){	// kray calls this function when rendering process end (render finished or interupted);
		MessageBox(hWnd,TEXT("Render interrupted"),TEXT("Info"),MB_ICONINFORMATION|MB_SETFOREGROUND);
		
		PostMessage(hWnd,WM_DESTROY,0,0);
	}

// interface, text operations
	void textLine(void* streamHandle,const char* txt){
		if (isEventString(KRAY_EVENT_RENDER_FINISHED,txt)){
			renderFinished();
		}else if (isEventString(KRAY_EVENT_RENDER_INTERRUPTED,txt)){
			renderInterupted();
		}else if (windowCreated){
			SetWindowTextA(hWnd,txt);
		}
	}

	int getTextFlags(void* handle){	
		return isInterrupted?KRAY_TEXT_GET_INTERRUPT:0;		// return interupt flag to Kray when user wants to stop rendering
	}

	void setPixels(void* windowHandle,const KrayEventPixelTable& kept){	// kray calls this funtion to set preview pixels
		EnterCriticalSection(&criticalSection);

		HDC hdc=GetDC(hWnd);	
		HDC hdcMem=CreateCompatibleDC(hdc);
		SelectObject(hdcMem,bitmap);

		for (size_t a=0 ; a<kept.count ; a++){
			int c=kept.pxls[a].c;

			SetPixel(hdcMem,kept.pxls[a].x,kept.pxls[a].y,RGB((c>>16)&0xff,(c>>8)&0xff,c&0xff));
		}

		ReleaseDC(hWnd,hdc);
		DeleteDC(hdcMem);

		LeaveCriticalSection(&criticalSection);
	}
	void copyPixels(void* windowHandle,const KrayEventPixelRectangle& kepr){	// kray calls this when it needs to modify bunch of pixels at once, no need to implement this, but its faster then calling series of setPixel() in its default implementation
		EnterCriticalSection(&criticalSection);
		
		HDC hdc=GetDC(hWnd);	
		HDC hdcMem=CreateCompatibleDC(hdc);
		SelectObject(hdcMem,bitmap);

		if (1){
			// quick copy version using StretchDIBits()

			BITMAPINFOHEADER DIBits;
			DIBits.biSize=sizeof(BITMAPINFOHEADER);
			DIBits.biPlanes=1;
			DIBits.biBitCount=32;
			DIBits.biCompression=BI_RGB;

			int xs=(kepr.x2-kepr.x1);
			int ys=(kepr.y2-kepr.y1)+1;
			DIBits.biWidth=kepr.modulo;
			DIBits.biHeight=ys;
			StretchDIBits(hdcMem,kepr.x1,kepr.y1,xs,ys,0,ys,xs,-ys,kepr.pixels,(BITMAPINFO*)&DIBits,DIB_RGB_COLORS,SRCCOPY);
		}else{
			// slower pixel by pixel version of the same routine (slower, but easier to understand, disabled with if)
			unsigned int* pixels=kepr.pixels;
			int modulo=kepr.modulo;

			pixels+=kepr.y1*modulo+kepr.x1;
			modulo-=(kepr.x2-kepr.x1);
			for (int y=kepr.y1 ; y<kepr.y2 ; y++){
				for (int x=kepr.x1 ; x<kepr.x2 ; x++){
					unsigned int c=*pixels++;
					SetPixel(hdcMem,x,y,RGB((c>>16)&0xff,(c>>8)&0xff,c&0xff));
				}
				pixels+=modulo;
			}
		}

		ReleaseDC(hWnd,hdc);
		DeleteDC(hdcMem);

		LeaveCriticalSection(&criticalSection);		
	}
	int setWindowSize(void* windowHandle,const KrayEventWindow& kew){	// kray calls this function to report preview size
		previewSizeX=kew.width;
		previewSizeY=kew.height;
		
		EnterCriticalSection(&criticalSection);
		freeBitmap();	// free previous bitmap
		
		HDC hdc=GetDC(hWnd);	
		bitmap=CreateCompatibleBitmap(hdc,kew.width,kew.height);
		ReleaseDC(hWnd,hdc);

		LeaveCriticalSection(&criticalSection);
		
		InvalidateRect(hWnd,0,FALSE);	// repaint window

		return 1;
	};
};