/*
	
	How to build a simple objects hierarchy, setup scene and render.
		by Grzegorz Tanski

*/

#include <iostream>

#include "../kraysdk/kray.h"
#include "../kraysdk/proto/direct.h"
#include "../kraysdk/type/coordsys.h"
#include "../kraysdk/type/vector.h"
#include "../kraysdk/type/axes.h"
#include "../kraysdk/symbol/mesh.h"
#include "../kraysdk/math/matrix.h"

#include "../SharedSources/basicMesh.h"
#include "../SharedSources/libraryDir.h"

#include <math.h>

#ifndef PI 
#define PI 3.1415926535897932384626433832795
#endif

int kraySdkExampleMain(int argc, const char** argv){	// main function is in startup.cpp
	Kray::Library klib(KRAY_DIRECTORY);				// here we start with KrayLib object

	Kray::Instance* kinst=klib.createInstance();	// then we create Kray instance, each instance have its own scene and settings
	
	if (kinst){										// check if instance was created, if not, Kray library wasn't found
		{
			Kray::DirectPrototyper prot(kinst);		// prototyper gives access to Kray Script command prototypes

			Kray::MeshSymbol msh(prot);				// create a mesh
			buildBoxMesh(prot,msh);					// call box build function
			setupSimpleMaterialsForMesh(prot,msh);	// call material creating function

			prot.sett_instanceRecurse(5);			// we need to set instance recursion limit before adding instances

			prot.group_selectInstance("box");
			prot.group_attachSet_mesh(msh,Kray::CoordSys(),0);	// add mesh to scene with given position and orientation

			prot.group_selectInstance("sphere");

			{
				const int dens=40;

				int c=0;

				for (int y=0 ; y<=dens ; y++){
					double ysin=sin(((double)y+0.5)*PI/dens);
					double ycos=cos(((double)y+0.5)*PI/dens);

					int xdens=(int)(ysin*dens*2);

					for (int x=0 ; x<xdens ; x++){
						double xsin=sin((double)x*2*PI/xdens);
						double xcos=cos((double)x*2*PI/xdens);
				
						Kray::Vector pos(xcos*ysin,xsin*ysin,ycos);

						prot.instanceAdd("box",Kray::CoordSys(pos,Kray::AxesHpb(c/33,c/41,c/23).scale(1.0/dens)));

						c++;
					}
				}
			}

			prot.group_select("waves");
			{
				const int dens=70;

				for (int y=-dens ; y<=dens ; y++){
					for (int x=-dens ; x<=dens ; x++){
						double q=(double)(x*x+y*y)/(dens*dens);

						Kray::Vector pos((double)x/dens,-0.4*exp(-q*10),(double)y/dens);
					
						prot.instanceAdd("sphere",Kray::CoordSys(pos,Kray::AxesHpb(x,y,q).scale(1.0/dens)));
					}
				}
			}

			prot.camera_perspective(Kray::Vector(800,600),400,Kray::CoordSys(Kray::Vector(-1,-0.05,-0.9),Kray::AxesHpb(0.8,-0.1,0)));

			prot.group_select("lights");

			prot.lightAdd_point(Kray::Vector(-5,-3,2),Kray::Vector(0.9,0.9,0.9));		// add light source to scene
//			prot.lightAdd_point(Kray::Vector(-4,-7,6),Kray::Vector(0.6,0.2,0.1));		// add light source to scene
//			prot.lightAdd_point(Kray::Vector(-2,-8,3),Kray::Vector(0.1,0.2,0.6));		// add light source to scene
			prot.lightAdd_ambient(Kray::Vector(0.2,0.2,0.2));

			prot.background_sky(); // gradient background

			prot.showInfo_scene();
			prot.sett_imageSampler_totalRotUniform(3);
			prot.sett_pixelFilter_mitchell();
			prot.pv_flayer(0,0.1,0);
		}

		kinst->parse("interactive;");	// interactive is outside {} so all objects we use in our core are unlinked, user can freely call 'reset' without risk of deleting objects which C++ code holds references to
	}else{
		std::cout << "Failed to create Kray instance" << std::endl;
	}

	return 0;
}
