#ifndef SDK_DIRECT_RAYTRACE_H
#define SDK_DIRECT_RAYTRACE_H

#include "calling.h"
#include "extray.h"
#include "object.h"

class KrayLibraryInstance;

#ifdef __cplusplus

namespace Kray{

typedef kl_ray_input RaytraceInput;
typedef kl_ray_output RaytraceOutput;

struct RaytraceBufferDimensions{
	size_t bucketSize[3];	// size of single bucket
	size_t buckets[3];		// number of buckets
	size_t allocSize[3];	// maximum size of job
};

struct RaytraceFunctions{
	static void* KRAYDLL_CALL_TYPE allocRaytraceBuffer(class KrayLibraryInstance* kli);
	static void KRAYDLL_CALL_TYPE freeRaytraceBuffer(void* buffer);
	static char KRAYDLL_CALL_TYPE setRaytraceBufferSize(void* buffer,const size_t maxJobSize[3]);	// returns 0 when failed
	
	static void KRAYDLL_CALL_TYPE getDimensions(void* buffer,struct RaytraceBufferDimensions* d);
	static void* KRAYDLL_CALL_TYPE getRaytraceBucket(void* buffer,const size_t index,int type);
	
	static void KRAYDLL_CALL_TYPE raytrace(void* buffer,const size_t job[3]);
};

class RaytraceFunctionsVtable : public Object{
public:
	void* (KRAYDLL_CALL_TYPE* allocRaytraceBuffer)(class KrayLibraryInstance* kli);
	void (KRAYDLL_CALL_TYPE* freeRaytraceBuffer)(void* buffer);
	char (KRAYDLL_CALL_TYPE* setRaytraceBufferSize)(void* buffer,const size_t maxJobSize[3]);	// returns 0 when failed

	void (KRAYDLL_CALL_TYPE* getDimensions)(void* buffer,struct RaytraceBufferDimensions* d);
	void* (KRAYDLL_CALL_TYPE* getRaytraceBucket)(void* buffer,const size_t index,int type);
	
	void (KRAYDLL_CALL_TYPE* raytrace)(void* buffer,const size_t job[3]);

	void init(){
		allocRaytraceBuffer=&(RaytraceFunctions::allocRaytraceBuffer);
		freeRaytraceBuffer=&(RaytraceFunctions::freeRaytraceBuffer);
		setRaytraceBufferSize=&(RaytraceFunctions::setRaytraceBufferSize);

		getDimensions=&(RaytraceFunctions::getDimensions);
		getRaytraceBucket=&(RaytraceFunctions::getRaytraceBucket);

		raytrace=&(RaytraceFunctions::raytrace);
	}	
};

class RaytraceBufferBase : public Object, public RaytraceBufferDimensions{
public:
	size_t totalBuckets(){
		return buckets[0]*buckets[1]*buckets[2];
	}
	const RaytraceBufferDimensions& dimensions() const{
		return *this;
	}
	void setDimensions(const RaytraceBufferDimensions& a){
		(*(RaytraceBufferDimensions*)this)=a;
	}
	
	static void fromIndex(size_t index,size_t bucket[3],const size_t buckets[3]){
		bucket[0]=index%buckets[0];
		index/=buckets[0];
		
		bucket[1]=index%buckets[1];
		index/=buckets[1];
		
		bucket[2]=index;
	}
	static size_t toIndex(const size_t bucket[3],const size_t buckets[3]){
		return bucket[0]+buckets[0]*(bucket[1]+buckets[1]*bucket[2]);
	}

	void indexToBucket(size_t index,size_t bucket[3]) const{
		fromIndex(index,bucket,buckets);
	}
	size_t bucketToIndex(const size_t bucket[3]) const{
		return toIndex(bucket,buckets);
	}

	void countBucketShift(size_t shift[3],const size_t bucket[3]){
		for (int a=0 ; a<3 ; a++){
			shift[a]=bucket[a]*bucketSize[a];
		}
	}
	void countBucketSize(size_t size[3],const size_t bucket[3],const size_t limit[3]) const{
		for (int s=0 ; s<3 ; s++){
			size_t shift=bucket[s]*bucketSize[s];

			size[s]=shift+bucketSize[s];
			
			if (size[s]>limit[s]){
				size[s]=limit[s];
			}
			
			if (size[s]>shift){
				size[s]-=shift;
			}else{
				size[s]=0;
			}
		}		
	}
};

template <class T> class RaytraceBufferTemplate : public RaytraceBufferBase{
	friend class RaytraceAccess;

protected:
	T** data;

public:
	RaytraceBufferTemplate():data(0){
	}
	~RaytraceBufferTemplate(){
		free();
	}
	
	T* getBucketBuffer(const size_t bucketIndex) const{
		return data[bucketIndex];
	}
	T* getBucketBuffer(const size_t bucket[3]) const{
		return data[bucketToIndex(bucket)];
	}

	void alloc(){
		free();
		data=new T*[totalBuckets()];
	}
	void free(){
		delete[] data;
		data=0;
	}

	T*& getData(size_t bucket){
		return data[bucket];
	}

	T& operator[](size_t index) const{	// get job number {index,0,0}
		return data[index/bucketSize[0]][index%bucketSize[0]];
	}
	
	template <class P> void process(P& processor);
};

class RaytraceInputBuffer : public RaytraceBufferTemplate<RaytraceInput>{
public:
	typedef RaytraceInput CellType;
};

class RaytraceOutputBuffer : public RaytraceBufferTemplate<RaytraceOutput>{
public:
	typedef RaytraceOutput CellType;
};

class RaytraceAccess : public Object{
	friend class Instance;
	
	RaytraceFunctionsVtable* krt;
	void* rtBuf;
	
	RaytraceInputBuffer input;
	RaytraceOutputBuffer output;

	RaytraceAccess(RaytraceFunctionsVtable* krt,KrayLibraryInstance* kli);
public:
	~RaytraceAccess();
	
	char allocBuffer(const size_t jobSize[3]);
	
	const RaytraceInputBuffer& getInput() const{
		return input;
	}
	const RaytraceOutputBuffer& getOutput() const{
		return output;
	}

	void raytrace(const size_t job[3]){
		krt->raytrace(rtBuf,job);
	}
};

};

#endif

#endif