#ifndef SDK_KRAY_H
#define SDK_KRAY_H

/*

	Kray SDK start header

	For more information go to http://www.kraytracing.com/wiki/Kray3_SDK

*/

#ifndef __cplusplus
#error C++ is required
#endif

#include "base/commons.h"
#include "base/calling.h"
#include "base/corevtable.h"
#include "base/prscall.h"
#include "base/object.h"

#ifdef _WINDOWS_LIKE_DLL
#include <windows.h>	// only for HINSTANCE hLib
#endif

#ifdef _UNIX_LIKE_DLL
#include <dlfcn.h>
#endif

class KrayLibraryPort;

namespace Kray{

class Instance;
class EventListener;

typedef void (KRAYDLL_CALL_TYPE *EventHandler)(class EventData*);
typedef void (*KrayLibInitFunctionPointer)(class Library*);

class Vtables : public Object{
public:
	class CoreFunctionsVtable* kcore;
	class MTFunctionsVtable* kmt;
	class BitmapFunctionsVtable* kbmp;
	class RaytraceFunctionsVtable* krt;
	class ParserFunctionsVtable* kprs;

public:
	Vtables():kcore(0),kmt(0),kbmp(0),krt(0),kprs(0){
	}

	char init();
	void exit();

	char ok() const{
		return kcore!=0;
	}
};

class Library : public Object{
	friend class Instance;
	friend class LibraryLoader;
	friend class StaticLibraryLoader;
	friend class ::KrayLibraryPort;

public:
	static const int flagUnloadIfEmpty=1;
	static const int flagLibraryIsBeingDestructed=2;
	static const int flagDontReleaseInstances=4;		// do not destroy unrelease instances when Library is destroyed
	
	static const int flagDefault=flagUnloadIfEmpty;

	void* krayModules;					// internal library linking data

private:
	void* libraryUserData;

	const char* libraryDir;
	Vtables vts;
		
	void* hLib;	// handle to system library object (platform dependent)
	
	Instance* firstInstance;

	int hostApiVersion;
	int flags;

public:
	Library(const char* libraryDirWithTailingSlashOrNull=0);
	~Library();

	void setFlag(int flag){
		flags|=flag;
	}
	void clrFlag(int flag){
		flags&=~flag;
	}
	int getFlag(int flagOrMask) const{
		return flags&flagOrMask;
	}

	void* getUserData() const{
		return libraryUserData;
	}
	void setUserData(void* ud){
		libraryUserData=ud;
	}

	bool load(const char* libraryDir=0);
	void release();				// unloads library (destroys all created instances)

	bool hasInstances() const;	// returns true if there are still allocated instances for this library

	bool isLoaded() const;		// is Kray library correctly loaded?

	class Instance* createInstance();	// instance is created with 'new' operator and can be freed using 'delete'
	Instance* createInstanceWithListner(EventListener* listner);
	Instance* createInstanceWithHandler(EventHandler handler,void* userData=0);

	void releaseInstance(Instance* inst);	// one may also use kray script 'delete inst;' instead of calling this function
	
	int getLibApiVersion() const;
	int getHostApiVersion() const{
		return hostApiVersion;
	}
	const char* getInfoString(int mode) const;
	int setPlatformOption(const char* option,const char* stringData,void* binaryData) const;
	const char* getLibraryDir() const;
	int getInfoInt(int mode) const;

	const MTFunctionsVtable* vtableMT() const{
		return vtable().kmt;
	}
	const Vtables& vtable() const{
		return vts;
	}

protected:
	Library(KrayLibInitFunctionPointer dlli,const char* dir);
	void zero();

	void unregister(Instance*);
	bool initLib(void* hLib);						// find init procedure adress and call
	bool initLib(KrayLibInitFunctionPointer dlli);	// call init procedure

	Instance* allocInstance();
	Instance* initInstance(Instance* i);

	void setLibraryDir(const char* dir);	// libraryDir must be ended with slash, its simply concatenated with 'kray' lib name
											// so "/dir/subdir/", "d:\dir\subdir\", "..\dir\" is ok, but "/dir/subdir", "d:\dir\subdir" "..\dir" is not
											// one may use platform independent DLLCURRENTDIR DLLROOTDIR DLLSLASH defined in "utils/dl.h"
};

class Instance : public Object{
	friend struct CoreFunctions;
	friend struct MTFunctions;
	friend class Library;
	
	Library* parent;
	Instance* nextInstance;
	
	const CoreFunctionsVtable* kcore;	// copy of &parent->kcore

	KrayLibraryInstance* kinst;		// lib communication end
	KrayScriptParser* kparser;		// default instance parser
	
	EventListener* defaultListener;	// default message handling object

	EventHandler handler;
	EventListener* listenerObject;
	void* userData;
	
private:
	Instance(Library* parent);

	const CoreFunctionsVtable* vtableCore() const{
		return kcore;
	}

	void setDefaultEventListener();
	void setEventListener(EventListener* eventListener=0);
	void setEventHandler(EventHandler handler,void* userData=0);

public:
	EventListener* getEventListener() const;

public:
	~Instance();
	
// inlined functions
	Library* getKrayLib() const{
		return parent;
	}
	int addPolyToMesh(const class PolyAddInfo& kpai) const{
		return vtableCore()->addPolyToMesh(kpai);
	}
	KrayObject addTriangleDirect(const class TriangleAddInfo& ktai) const{
		return vtableCore()->addTriangleDirect(kinst,ktai);
	}

	EventListener* getListener() const{	
		return listenerObject;
	}
	void* getUserData() const{
		return userData;
	}
	KrayLibraryInstance* getKrayInstanceHandle() const{
		return kinst;
	}
	void callHandler(class EventData* data){
		handler(data);
	}

	KrayHandle getObjectByName(const char* name,KrayHandleType type,KraySymbolHandle* symbolHandle,const char* context=0) const{
		return vtableParser()->getObjectByName(kparser,name,symbolHandle,type,context);
	}
	KrayHandle getObjectByName(const char* name,KrayHandleType type,const char* context=0) const{
		KraySymbolHandle symbolHandle;
		
		return vtableParser()->getObjectByName(kparser,name,&symbolHandle,type,context);
	}
	
	void getSceneInfo(SceneInfo* ksi){
		asyncWaitParse();
		vtableCore()->getSceneInfo(kinst,ksi);
	}
	
	int parse(const char* script,const char* name=0) const{			// regular parser
		return vtableParser()->parseScript(kparser,script,name,0);
	}
	int parseQuick(const char* script,const char* name=0) const{	// quick parser (no expressions, all values explicit)
		return vtableParser()->parseScript(kparser,script,name,1);
	}
	int parseBinary(const char* script) const{						// binary packed script (faster access to Kray API)
		return vtableParser()->parseScript(kparser,script,0,2);
	}

	void asyncParse(bool asyncModeOn){
		vtableParser()->parseAsync(kparser,asyncModeOn?KRAY_ASYNC_PARSE_ON:KRAY_ASYNC_PARSE_OFF,0);
	}
	int asyncWaitParse(){
		return vtableParser()->parseAsync(kparser,KRAY_ASYNC_PARSE_WAIT,0);
	}
	void asyncFlushParse(){
		vtableParser()->parseAsync(kparser,KRAY_ASYNC_PARSE_FORGET,0);
	}
	void asyncSetBufferSize(size_t size){
		vtableParser()->parseAsync(kparser,KRAY_ASYNC_PARSE_SET_SIZE,size);		
	}

	KrayScriptParser* getDefaultParser() const{
		return kparser;
	}

	const MTFunctionsVtable* vtableMT() const{
		return parent->vtable().kmt;
	}
	const BitmapFunctionsVtable* vtableBitmap() const{
		return parent->vtable().kbmp;
	}
	const ParserFunctionsVtable* vtableParser() const{
		return parent->vtable().kprs;
	}

	class RaytraceAccess* createRaytraceAccess() const;

	void sendBackEvent(EventData& ked) const{
		vtableCore()->processMessage(kinst,ked);
	}
};

};

#endif
