#ifndef KRAY_BITMAPSYMBOL_H
#define KRAY_BITMAPSYMBOL_H

#include "../base/parser.h"
#include "../base/bmp.h"
#include "symbol.h"

namespace Kray{

class Bitmap : protected KrayBitmapInfo{
	friend class BitmapSymbol;

	union IntCharUnion{
		int icols;
		unsigned char cols[4];
	};

	template <class T> static T avg(const T* rgba){
		return (rgba[0]+rgba[1]+rgba[2])/3;
	}
	template <class T> static T clip(T v){
		if (v<0) v=0;
		if (v>1) v=1;

		return v;
	}
public:
	Bitmap(){
		pixels=0;
		sizx=0;
		sizy=0;
	}
	Bitmap(void* pixels,KrayPixelFormat format,int modulo){
		this->pixels=pixels;
		this->pixelFormat=format;
		this->sizx=modulo;
		this->sizy=0;		// unknown
	}

	int width() const{
		return sizx;
	}
	int height() const{
		return sizy;
	}
	char ok() const{
		return pixels!=0;
	}

	/////////////////////////////////////////////////////////////

	void setChar1(int x,int y,unsigned char rgba) const{
		((unsigned char*)pixels)[width()*y+x]=rgba;
	}
	void setChar4(int x,int y,int rgba) const{	// rgba must be aligned to 4, sizeof(int) must be 4
		((int*)pixels)[width()*y+x]=rgba;
	}

	template <class T> void setFloat1(int x,int y,T value) const{
		((float*)pixels)[width()*y+x]=(float)value;
	}
	template <class T> void setFloat3(int x,int y,const T rgb[3]) const{
		float* s=(float*)pixels;
		s+=(y*width()+x)*3;
		
		for (int i=0 ; i<3 ; i++){
			s[i]=(float)rgb[i];
		}
	}
	template <class T> void setFloat3x(int x,int y,const T rgb[3]) const{
		float* s=(float*)pixels;
		s+=(y*width()+x)*4;
		
		for (int i=0 ; i<3 ; i++){
			s[i]=(float)rgb[i];
		}
	}
	template <class T> void setFloat3a(int x,int y,const T rgb[3],T a) const{
		float *s=(float*)pixels;
		s+=(y*width()+x)*4;
		
		for (int i=0 ; i<3 ; i++){
			s[i]=(float)rgb[i];
		}
		s[3]=(float)a;
	}
	template <class T> void setFloat4(int x,int y,const T rgba[3]) const{
		float *s=(float*)pixels;
		s+=(y*width()+x)*4;
		
		for (int i=0 ; i<4 ; i++){
			s[i]=(float)rgba[i];
		}
	}	

	/////////////////////////////////////////////////////////////

	unsigned char getChar1(int x,int y) const{
		return ((unsigned char*)pixels)[width()*y+x];	
	}
	void getChar4(int x,int y,int& rgba) const{	// rgba must be aligned to 4, sizeof(int) must be 4
		rgba=((int*)pixels)[width()*y+x];		
	}

	template <class T> void getFloat1(int x,int y,T& value) const{
		value=(T)((float*)pixels)[width()*y+x];		
	}
	template <class T> void getFloat3(int x,int y,T rgb[3]) const{
		float *s=(float*)pixels;
		s+=(y*width()+x)*3;
		
		for (int i=0 ; i<3 ; i++){
			rgb[i]=(T)s[i];
		}
	}
	template <class T> void getFloat3x(int x,int y,T rgb[3]) const{
		getFloat3a(x,y,rgb);
	}
	template <class T> T getFloat3a(int x,int y,T rgb[3]) const{
		float *s=(float*)pixels;
		s+=(y*width()+x)*4;
		
		for (int i=0 ; i<3 ; i++){
			rgb[i]=(T)s[i];
		}

		return (T)s[3];
	}
	template <class T> void getFloat4(int x,int y,T rgba[3]) const{
		float *s=(float*)pixels;
		s+=(y*width()+x)*4;
		
		for (int i=0 ; i<4 ; i++){
			rgba[i]=(T)s[i];
		}
	}

	/////////////////////////////////////////////////////////////
	
	template <class T> void setPixel(int x,int y,const T* rgb,T a,KrayPixelFormat pixelFormat) const{
		switch(pixelFormat){
			case PF_CHAR_GRAY:{
				setChar1(x,y,(unsigned char)(clip(avg(rgb))*255));
			}break;
			case PF_CHAR_RGBA:{
				IntCharUnion u;
			
				for (int i=0 ; i<3 ; i++){
					u.cols[i]=(unsigned char)(clip(rgb[i])*255);
				}
				u.cols[3]=(unsigned char)(clip(a)*255);
			
				setChar4(x,y,u.icols);
			}break;
			case PF_FLOAT_GRAY:
				setFloat1(x,y,avg(rgb));
			break;
			case PF_FLOAT_RGB:
				setFloat3(x,y,rgb);
			break;
			case PF_FLOAT_RGBx:
				setFloat3x(x,y,rgb);
			break;
			case PF_FLOAT_RGBA:
				setFloat3a(x,y,rgb,a);
			break;

			default:
			break;
		}
	}

	//////////////////////////////////////////////////////////

	template <class T> int getPixel(int x,int y,T* rgba,KrayPixelFormat pixelFormat) const{
		switch(pixelFormat){
			case PF_CHAR_GRAY:{
				rgba[0]=(T)(getChar1(x,y)/255.0);
				rgba[1]=rgba[0];
				rgba[2]=rgba[0];

				return 3;
			}break;		
			case PF_CHAR_RGBA:{
				IntCharUnion u;

				getChar4(x,y,u.icols);
			
				for (int i=0 ; i<4 ; i++){
					rgba[i]=(T)(u.cols[i]/255.0);
				}
			
				return 4;
			}break;		
			case PF_FLOAT_GRAY:
				getFloat1(x,y,rgba[0]);
				rgba[1]=rgba[0];
				rgba[2]=rgba[0];
				
				return 3;
			break;
			case PF_FLOAT_RGB:
				getFloat3(x,y,rgba);
				
				return 3;
			break;
			case PF_FLOAT_RGBx:
				getFloat3x(x,y,rgba);
				
				return 3;
			break;
			case PF_FLOAT_RGBA:
				getFloat4(x,y,rgba);
				
				return 4;
			break;
			default:
			break;
		}

		return 0;
	}

	/////////////////////////////////////////////////////////////////

	template <class T,KrayPixelFormat PF> void setPixel(int x,int y,const T* rgba) const{
		setPixel(x,y,rgba,rgba[3],PF);
	}
	template <class T,KrayPixelFormat PF> void setPixel(int x,int y,const T* rgb,T a) const{
		setPixel(x,y,rgb,a,PF);
	}
	template <class T> void setPixel(int x,int y,const T* rgba) const{
		setPixel(x,y,rgba,rgba[3],pixelFormat);
	}
	template <class T> void setPixel(int x,int y,const T* rgb,T a) const{
		setPixel(x,y,rgb,a,pixelFormat);
	}

	/////////////////////////////////////////////////////////////////

	template <class T,KrayPixelFormat PF> int getPixel(int x,int y,T* rgba) const{
		return getPixel(x,y,rgba,PF);
	}
	template <class T> int getPixel(int x,int y,T* rgba) const{
		return getPixel(x,y,rgba,pixelFormat);
	}

	/////////////////////////////////////////////////////////////////

	void setPixel(int x,int y,const class Vector& rgba) const;
	void getPixel(int x,int y,Vector& rgba) const;

	/////////////////////////////////////////////////////////////////

	void copyTo(const Bitmap* bmp,int width,int height) const{	// both this and bmp must have the same dimensions, but can have different modulo (row length)
		float rgba[4];
		
		rgba[3]=1;	// default alpha in case there is no alpha in image

		for (int y=0 ; y<height ; y++){
			for (int x=0 ; x<width ; x++){
				getPixel(x,y,rgba);
				bmp->setPixel(x,y,rgba);
			}
		}
	}
};

class BitmapSymbol : public Symbol{
	Bitmap kbo;

	void registerSymbol(int width,int height,KrayPixelFormat pixelFormat);
	char setSharedMemory(void* mem,int width,int height);

public:
	// creates new bitmap with given parameters
	BitmapSymbol(CommandBuffer& ktb,int width,int height,KrayPixelFormat pixelFormat=PF_FLOAT_RGBA,const char* name=0);

	// creates new bitmap with given parameters and already allocated memory
	BitmapSymbol(CommandBuffer& ktb,void* memory,int width,int height,KrayPixelFormat pixelFormat=PF_FLOAT_RGBA,const char* name=0);

	// creates empty bitmap of zero size
	BitmapSymbol(CommandBuffer& ktb,const char* name=0);	

	// makes a copy of bitmap and register in Kray
	BitmapSymbol(CommandBuffer& ktb,const Bitmap& bitmap,int width,int height,const char* name=0);

	// loads a bitmap via Kray's bitmap cache (cache creates references to already loaded files instead of copies)
	BitmapSymbol(const char* fileName,CommandBuffer& ktb,const char* name=0);			
	
public:
	void refreshDirectLink();

	const KrayBitmapInfo& getInfo(){
		refreshDirectLink();

		return kbo;
	}
	Bitmap* getDirect(){
		refreshDirectLink();

		if (!kbo.ok()){
			return 0;
		}

		return &kbo;
	}

	int directWidth(){
		refreshDirectLink();

		return kbo.height();
	}
	int diredtHeight(){
		refreshDirectLink();

		return kbo.width();
	}
	char directOk(){
		refreshDirectLink();

		return kbo.ok();
	}

	void setup(int width,int height,KrayPixelFormat pixelFormat);
	void setupSharedMemory(void* memory,int width,int height,KrayPixelFormat pixelFormat);

	void slowDirectSetPixel(int x,int y,const Vector& rgba){	// those are slow indeed, faster way is to getDirect() and then setPixel()
		refreshDirectLink();

		kbo.setPixel(x,y,rgba);
	}
	void slowDirectGetPixel(int x,int y,Vector& rgba){
		refreshDirectLink();

		kbo.getPixel(x,y,rgba);
	}

	void plot(int x,int y,const Vector& rgba){
		setPixel(x,y,rgba);
	}
	void setPixel(int x,int y,const Vector& rgba);
};

};

#endif