#include <stddef.h>

#include <math.h>

#include "../kraysdk/kcmds.h"
#include "../kraysdk/kray.h"
#include "../kraysdk/base/prscall.h"
#include "../kraysdk/base/bmpcall.h"
#include "../kraysdk/base/vmap.h"
#include "../kraysdk/buffer/dynamic.h"

#include "../kraysdk/symbol/symbol.h"
#include "../kraysdk/symbol/bitmap.h"
#include "../kraysdk/symbol/vertexmap.h"
#include "../kraysdk/symbol/mesh.h"

#include "../kraysdk/type/double.h"
#include "../kraysdk/type/string.h"
#include "../kraysdk/type/vector.h"
#include "../kraysdk/type/axes.h"
#include "../kraysdk/type/coordsys.h"
#include "../kraysdk/math/matrix.h"

#include "../kraysdk/buffer/text.h"

#include "../kraysdk/base/getinfo.h"
#include "../kraysdk/proto/proto.h"

using namespace Kray;

Symbol::Symbol(CommandBuffer& ktb,const char* name){
	initSymbol(ktb,name);	
}
void Symbol::initSymbol(CommandBuffer& ktb,const char* name){
	commandBufferHandle=&ktb;
	directObjectHandle=0;
	directSymbolHandle=0;

	if (name){
		setName(name);
	}else{
		generateAutoname();
	}	
}
void Symbol::setName(const char* name){
	int a=0;
	for ( ; name[a] && a<(kraySymbolNameMaxChars-1) ; a++){
		this->name[a]=name[a];
	}
	
	this->name[a]=0;
}
void Symbol::generateAutoname(){
	ktb().generateAutoname(*this,name);
}
void Symbol::registerSymbol(){
	registerSymbol(KraySymbolType);
}
bool Symbol::isRegistered() const{
	return ktb().getKrayInstance()?validSymbol():1;
}
char* Symbol::toString(DynamicBuffer& rb){
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=KrayDtStringableType;
	kpeb.u.stringable.string=rb.getBuffer();
	kpeb.u.stringable.size=rb.size();
		
	getValue(kpeb);

	if (!kpeb.u.stringable.string && kpeb.u.stringable.size){
		rb.fit(kpeb.u.stringable.size+1);

		kpeb.u.stringable.string=rb.getBuffer();
		kpeb.u.stringable.size=rb.size();

		getValue(kpeb);
	}

	return kpeb.u.stringable.string;
}
void Symbol::registerSymbol(KrayHandleType type,const char* context){
	if (ktb().getKrayInstance()){
		directObjectHandle=ktb().getKrayInstance()->getObjectByName(name,type,&directSymbolHandle,context);
	}else{
		directObjectHandle=0;
	}

	updateSymbol();
}
void Symbol::freeSymbol(){
	if (isRegistered()){
		(ktb() + KrayCmd(delete) + (*this)).parse();
	}
}

void Symbol::getValue(KrayParserExchangeBuffer& kpeb) const{
	ktb().getKrayInstance()->vtableParser()->getValue(ktb().getParser(),directObjectHandle,&kpeb);
}
void Symbol::setValue(const KrayParserExchangeBuffer& kpeb) const{
	ktb().getKrayInstance()->vtableParser()->setValue(ktb().getParser(),directObjectHandle,&kpeb);
}	

///////////////////////////////// Double

void DoubleSymbol::registerSymbol(){
	(ktb() + KrayCmd(var) + *((Symbol*)this)).parse();
	
	DatatypeSymbol::registerSymbol();
}
void DoubleSymbol::directSet(const Double& v){
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=v.dtType();
	kpeb.u.value=v;
		
	setValue(kpeb);
}
void DoubleSymbol::directGet(Double& v) const{
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=v.dtType();
	getValue(kpeb);

	if (kpeb.dtType==v.dtType()){
		v=kpeb.u.value;
	}else{
		ktb().error(PARSE_ERR_BAD_CAST);
		v=0;
	}
}

///////////////////////////////// String

void StringSymbol::registerSymbol(){
	(ktb() + KrayCmd(string) + *((Symbol*)this)).parse();
	
	DatatypeSymbol::registerSymbol();
}
void StringSymbol::directSet(const String& v){
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=v.dtType();
	kpeb.u.string=v;
		
	setValue(kpeb);
}
const char* StringSymbol::toString(DynamicBuffer& rb){
	String s;
	directGet(s);
	
	return s.string;
}
void StringSymbol::directGet(String& v) const{
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=v.dtType();
	kpeb.u.string.string=0;

	getValue(kpeb);

	if (kpeb.dtType==KrayDtStringType){
		v.set(kpeb.u.string.string);
	}else{
		ktb().error(PARSE_ERR_BAD_CAST);
		v.set();
	}
}

///////////////////////////////// Vector

Vector& Vector::operator*=(double m){
	scale(v,count,m);

	return (*this);
}
Vector& Vector::operator+=(const Vector& vc){
	if (vc.dim()==dim()){
		for (int a=0 ; a<dim() ; a++){
			v[a]+=vc.v[a];
		}
	}

	return (*this);
}
Vector& Vector::operator-=(const Vector& vc){
	if (vc.dim()==dim()){
		for (int a=0 ; a<dim() ; a++){
			v[a]-=vc.v[a];
		}
	}

	return (*this);
}
double Vector::length() const{
	return sqrt(dot(*this));
}
void Vector::cross(const Vector& v1,const Vector& v2){
	(*this)[0]=v1[1]*v2[2]-v1[2]*v2[1];
	(*this)[1]=v1[2]*v2[0]-v1[0]*v2[2];
	(*this)[2]=v1[0]*v2[1]-v1[1]*v2[0];	
	
	count=3;
}
double Vector::dot(const double* v1,const double* v2,int count){
	double res=0;

	for (int a=0 ; a<count ; a++){
		res+=v1[a]*v2[a];
	}

	return res;
}
void Vector::scale(double* v,int count,double value){
	for (int a=0 ; a<count ; a++){
		v[a]*=value;
	}
}
void VectorSymbol::registerSymbol(){
	(ktb() + KrayCmd(vector) + *((Symbol*)this)).parse();
	
	DatatypeSymbol::registerSymbol();
}
void VectorSymbol::directSet(const Vector& v){
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=v.dtType();
	kpeb.u.vector=v;
		
	setValue(kpeb);
}
void VectorSymbol::directGet(Vector& v) const{
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=v.dtType();
	getValue(kpeb);

	if (kpeb.dtType==v.dtType()){
		v=kpeb.u.vector;
	}else{
		ktb().error(PARSE_ERR_BAD_CAST);
		v.invalidate();
	}
}

///////////////////////////////// Axes

void Axes::asAxesHpb(const KrayAxesCore& core,AxesHpb& v){
	v.radians(core.matrix[0][0],core.matrix[0][1],core.matrix[0][2]);
	v.scale(core.matrix[1][0],core.matrix[1][1],core.matrix[1][2]);
}
void Axes::asMatrix(const KrayAxesCore& core,Matrix3x3& v){
	v.copy(core.matrix[0]);
}
bool Axes::isEye(const KrayAxesCore& core){
	if (core.count==0) return true;

	AxesHpb a;
	if (Axes(core).isAxesHpb(a)){
		return a.isEye();
	}

	Matrix3x3 m;
	if (Axes(core).isMatrix(m)){
		return m.isEye();
	}

	return 0;
}
bool Axes::isAxesHpb(AxesHpb& v) const{
	if (isAxesHpb()){
		asAxesHpb(v);

		return true;
	}

	return false;
}
bool Axes::isMatrix(Matrix3x3& v) const{
	if (isMatrix()){
		asMatrix(v);

		return true;
	}

	return false;
}
CoordSys Axes::toCoordSys() const{
	return toCoordSys(0,0,0);
}
CoordSys Axes::toCoordSys(double x,double y,double z) const{
	return CoordSys(Vector(x,y,z),*this);
}
void Axes::transform(KrayAxesCore& tt,char rev) const{
	Axes t(tt);

	Matrix3x3 mt;
	Matrix3x3 m0;
			
	t.toMatrix(mt);
	toMatrix(m0);

	if (rev){
		m0.inverse();
	}

	mt.multiply(m0);

	if (t.isAxesHpb() || isAxesHpb()){
		AxesHpb axt;
		AxesHpb ax0;

		t.toAxesHpb(axt);
		toAxesHpb(ax0);

		AxesHpb at;
		mt.toAxes(at);

		// fix scaling (better accuracy)
		for (int a=0 ; a<krayParserAxesMax ; a++){
			if (rev){
				at.scl(0)=axt.scl(0)/ax0.scl(0);
			}else{
				at.scl(0)=axt.scl(0)*ax0.scl(0);
			}
		}

		copy(tt,at);
	}else{
		t=mt;

		copy(tt,t);
	}
}
void Axes::toAxesHpb(const KrayAxesCore& core,AxesHpb& v){
	if (core.count==0){
		v.eye();
	}else if (core.count==3){
		Matrix3x3 m;

		toMatrix(core,m);

		m.toAxes(v);
	}else{
		asAxesHpb(core,v);
	}
}
void Axes::copy(KrayAxesCore& d,const KrayAxesCore& core){
	d.count=core.count;

	for (int b=0 ; b<d.count ; b++){
		for (int a=0 ; a<krayParserAxesMax ; a++){
			d.matrix[b][a]=core.matrix[b][a];
		}
	}
}
void Axes::toMatrix(const KrayAxesCore& core,Matrix3x3& v){
	if (core.count==0){
		v.eye();
	}else if (core.count==3){
		asMatrix(core,v);
	}else{
		AxesHpb a;

		asAxesHpb(core,a);

		a.toMatrix(v);
	}
}
Axes& Axes::scale(double v0,double v1,double v2){
	if (isMatrix()){
		for (int a=0 ; a<3 ; a++){
			matrix[a][0]*=v0;
			matrix[a][1]*=v1;
			matrix[a][2]*=v2;
		}
	}else if (isAxesHpb()){
		matrix[1][0]*=v0;
		matrix[1][1]*=v1;
		matrix[1][2]*=v2;
	}

	return (*this);
}
void Axes::getScale(double scl[3]) const{
	if (isMatrix()){
		for (int a=0 ; a<3 ; a++){
			scl[a]=sqrt(Vector::dot(matrix[a],matrix[a],3));
		}
	}else if (isAxesHpb()){
		for (int a=0 ; a<3 ; a++){
			scl[a]=matrix[1][a];
		}
	}else{
		for (int a=0 ; a<3 ; a++){
			scl[a]=1;
		}
	}
}

///////////////////////////////// AxesHpb

AxesHpb& AxesHpb::operator*=(const Matrix3x3& v){
	Matrix3x3 m(*this);

	m*=v;

	(*this)=m;

	return *this;
}
AxesHpb& AxesHpb::operator*=(const AxesHpb& v){
	return (*this)*=Matrix3x3(v);
}
Matrix3x3 AxesHpb::operator*(const AxesHpb& v) const{
	return (*this)*Matrix3x3(v);
}
Matrix3x3 AxesHpb::operator*(const class Matrix3x3& v) const{
	Matrix3x3 m(*this);
		
	m*=v;
		
	return m;
}
AxesHpb& AxesHpb::operator=(const Matrix3x3& m){
	m.toAxes(*this);

	return *this;
}

void AxesSymbol::registerSymbol(){
	(ktb() + KrayCmd(axes) + *((Symbol*)this)).parse();
	
	DatatypeSymbol::registerSymbol();
}
void AxesSymbol::directSet(const Axes& v){
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=v.dtType();
	kpeb.u.axes=v;
	
	setValue(kpeb);
}
void AxesSymbol::directGet(Axes& v) const{
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=v.dtType();
	getValue(kpeb);

	if (kpeb.dtType==v.dtType()){
		v=kpeb.u.axes;
	}else{
		ktb().error(PARSE_ERR_BAD_CAST);
		v.invalidate();
	}
}

/////////////////////////////////////// CoordSys
	
CoordSys::CoordSys(const Vector& v,const Axes& a){
	set(v);
	set(a);
}
CoordSys::CoordSys(const KrayCoordSysCore& core){
	Axes::copy(a,core.a);

	for (int a=0 ; a<krayParserAxesMax ; a++){
		v[a]=core.v[a];
	}
}
CoordSys::CoordSys(const Matrix4x4& v){
	set((Vector)v);
	set((Axes)v);
}

char CoordSys::transform(CoordSys& vv,char rev) const{
	Vector v;
	v.set(vv.v,3);

	if (!transform(v,rev)) return 0;

	v.copyTo(vv.v);

	Axes(a).transform(vv.a,rev);

	return 1;
}
char CoordSys::transform(Vector& vv,char rev) const{
	if (dim()!=krayParserAxesMax) return 0;

	Matrix3x3 m;
	Axes::toMatrix(a,m);

	if (rev){
		m.inverse();
		vv-=Vector::fromArray(v,3);
	}

	m.rotateVector(vv.v);

	if (!rev){
		vv+=Vector::fromArray(v,3);
	}

	return 1;
}
void CoordSys::clear(){
	Axes::clear(a);

	for (int a=0 ; a<krayParserAxesMax ; a++){
		v[a]=0;
	}
}
void CoordSys::set(const KrayAxesCore& ax){
	Axes::copy(a,ax);
}
Axes CoordSys::getAxes() const{
	return Axes(a);
}
Vector CoordSys::getVector() const{
	return Vector::fromArray(v,3);
}
void CoordSys::set(const KrayVectorCore& vec){
	if (vec.count!=krayParserAxesMax) return;

	for (int a=0 ; a<krayParserAxesMax ; a++){
		v[a]=vec.v[a];
	}
}
bool CoordSys::isCenteredAtZero() const{
	for (int a=0 ; a<krayParserAxesMax ; a++){
		if (v[a]!=0) return false;
	}

	return true;
}
bool CoordSys::isEyeAxes() const{
	return Axes::isEye(a);
}

///////////////

void CoordSysSymbol::directSet(const CoordSys& v){
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=v.dtType();
	kpeb.u.coordsys=v;
		
	setValue(kpeb);
}
void CoordSysSymbol::directGet(CoordSys& v) const{
	KrayParserExchangeBuffer kpeb;
	
	kpeb.dtType=v.dtType();
	getValue(kpeb);

	if (kpeb.dtType==v.dtType()){
		v=kpeb.u.coordsys;
	}else{
		ktb().error(PARSE_ERR_BAD_CAST);
		v.invalidate();
	}
}

/////////////////////////////////////// Matrix

Matrix3x3::Matrix3x3(double x0,double y0,double z0, double x1,double y1,double z1, double x2,double y2,double z2){
	(*this)[0][0]=x0;
	(*this)[0][1]=y0;
	(*this)[0][2]=z0;

	(*this)[1][0]=x1;
	(*this)[1][1]=y1;
	(*this)[1][2]=z1;

	(*this)[2][0]=x2;
	(*this)[2][1]=y2;
	(*this)[2][2]=z2;
}
bool Matrix3x3::isEye() const{
	for (int y=0 ; y<3 ; y++){
		for (int x=0 ; x<3 ; x++){
			double eyeVal=(x==y)?1.0:0.0;

			if ((*this)[y][x]!=eyeVal){
				return false;
			}
		}
	}

	return true;
}
bool Matrix3x3::isRightHanded() const{
	Vector xy;
	xy.cross(Vector().copy((*this)[0],3),Vector().copy((*this)[1],3));

	return Vector().copy((*this)[2],3).dot(xy)>0;
}
void Matrix3x3::inverse(){
	Matrix3x3 t;
	inverse(t);

	(*this)=t;
}
void Matrix3x3::inverse(Matrix3x3& r) const{
	const Matrix3x3& m=(*this);

	r[0][0]=m[1][1]*m[2][2]-m[1][2]*m[2][1];
	r[1][0]=m[1][2]*m[2][0]-m[1][0]*m[2][2];
	r[2][0]=m[1][0]*m[2][1]-m[1][1]*m[2][0];

	r[0][1]=m[0][2]*m[2][1]-m[0][1]*m[2][2];
	r[1][1]=m[0][0]*m[2][2]-m[0][2]*m[2][0];
	r[2][1]=m[0][1]*m[2][0]-m[0][0]*m[2][1];

	r[0][2]=m[0][1]*m[1][2]-m[0][2]*m[1][1];
	r[1][2]=m[0][2]*m[1][0]-m[0][0]*m[1][2];
	r[2][2]=m[0][0]*m[1][1]-m[0][1]*m[1][0];

	r.multiply(1.0/determinant());
}
void Matrix3x3::rotateVector(double v[3]) const{
	double c[3];

	for (int a=0 ; a<3 ; a++){
		c[a]=v[a];
	}

	const Matrix3x3& m=(*this);

	for (int i=0 ; i<3 ; i++){
		int j;

		for (j=0,v[i]=0 ; j<3 ; j++){
			v[i]+=m[j][i]*c[j];
		}
	}
}
double Matrix3x3::determinant() const{
	const Matrix3x3& m=(*this);

	return m[0][0]*m[1][1]*m[2][2]+
	       m[1][0]*m[2][1]*m[0][2]+
	       m[2][0]*m[0][1]*m[1][2]
	       -m[1][0]*m[0][1]*m[2][2]
	       -m[0][0]*m[2][1]*m[1][2]
	       -m[2][0]*m[1][1]*m[0][2];
}
void Matrix3x3::setRotation(int A,int B,double angle){
	eye();

	double st=sin(angle);
	double ct=cos(angle);

	(*this)[A][A]=ct;
	(*this)[A][B]=st;
	(*this)[B][A]=-st;
	(*this)[B][B]=ct;
}
void Matrix3x3::fromXYZ(double x,double y,double z){
	setRotation(0,1,z);

	Matrix3x3 ph;
	ph.setRotation(0,2,y);
	multiply(ph);

	ph.setRotation(1,2,x);
	multiply(ph);
}
void Matrix3x3::fromRPY(double x,double y,double z){
	setRotation(1,2,x);

	Matrix3x3 ph;
	ph.setRotation(0,2,y);
	multiply(ph);

	ph.setRotation(0,1,z);
	multiply(ph);
}
void Matrix3x3::fromAxes(const AxesHpb& a){
	setRotation(0,1,a.hpb(2));

/*	eye();

	double st=sin(a.hpb(2));
	double ct=cos(a.hpb(2));

	(*this)[0][0]=ct;
	(*this)[0][1]=st;
	(*this)[1][0]=-st;
	(*this)[1][1]=ct;*/
	
	Matrix3x3 ph;
	ph.setRotation(1,2,a.hpb(1));
/*	ph.eye();

	st=sin(a.hpb(1));
	ct=cos(a.hpb(1));
	
	ph[1][1]=ct;
	ph[1][2]=st;
	ph[2][1]=-st;
	ph[2][2]=ct;*/

	multiply(ph);

	ph.setRotation(2,0,a.hpb(0));
	
/*	ph[1][1]=1;
	ph[1][2]=0;
	ph[2][1]=0;

	st=sin(a.hpb(0));
	ct=cos(a.hpb(0));
	
	ph[2][2]=ct;
	ph[2][0]=st;
	ph[0][2]=-st;
	ph[0][0]=ct;*/

	multiply(ph);

	for (int i=0 ; i<3 ; i++){
		Vector::scale((*this)[i],3,a.scl(i));
	}
}
void Matrix3x3::toAxes(AxesHpb* ax,Matrix3x3& t){
	for (int a=0 ; a<3 ; a++){
		ax->scl(a)=sqrt(Vector::dot(t[a],t[a],3));
		
		Vector::scale(t[a],3,1.0/ax->scl(a));
	}

	if (!t.isRightHanded()){
		for (int a=0 ; a<3 ; a++){
			t[2][a]*=-1;
		}

		ax->scl(2)*=-1;
	}

	ax->hpb(0)=atan2(t[2][0],t[2][2]);
	
	double st=sin(ax->hpb(0));
	double ct=cos(ax->hpb(0));

	Matrix3x3 ph;
	ph.eye();
	ph[2][2]=ct;
	ph[2][0]=st;
	ph[0][2]=-st;
	ph[0][0]=ct;
		
	t.multiplyTransposed(ph);
	
	ax->hpb(1)=atan2(-t[2][1],t[2][2]);

	st=sin(ax->hpb(1));
	ct=cos(ax->hpb(1));
	
	ph[2][0]=0;
	ph[0][2]=0;
	ph[0][0]=1;
	
	ph[1][1]=ct;
	ph[1][2]=st;
	ph[2][1]=-st;
	ph[2][2]=ct;

	t.multiplyTransposed(ph);

	ax->hpb(2)=atan2(t[0][1],t[0][0]);	
}
void Matrix4x4::toMatrix3x3(Matrix3x3& t) const{
	for (int y=0 ; y<3 ; y++){
		for (int x=0 ; x<3 ; x++){
			t[y][x]=(*this)[y][x];
		}
	}
}
Matrix4x4::operator AxesHpb() const{
	Matrix3x3 t;

	toMatrix3x3(t);

	AxesHpb r;

	Matrix3x3::toAxes(&r,t);

	return r;
}
Matrix4x4::operator Axes() const{
	Matrix3x3 t;

	toMatrix3x3(t);

	return Axes(t);
}
Matrix4x4::operator Vector() const{
	double w=(*this)[3][3];

	return Vector((*this)[3][0]/w,(*this)[3][1]/w,(*this)[3][2]/w);
}

/////////////////////////////////////// Bitmap

void Bitmap::setPixel(int x,int y,const Vector& rgba) const{
	setPixel(x,y,rgba.v,rgba.count==4?rgba.v[3]:1,pixelFormat);
}
void Bitmap::getPixel(int x,int y,Vector& rgba) const{
	rgba.count=getPixel(x,y,rgba.v,pixelFormat);
}


void BitmapSymbol::setup(int width,int height,KrayPixelFormat pixelFormat){
	registerSymbol(width,height,pixelFormat);
	refreshDirectLink();
}
void BitmapSymbol::setupSharedMemory(void* memory,int width,int height,KrayPixelFormat pixelFormat){
	registerSymbol(0,0,pixelFormat);
	refreshDirectLink();
	setSharedMemory(memory,width,height);
}

BitmapSymbol::BitmapSymbol(CommandBuffer& ktb,int width,int height,KrayPixelFormat pixelFormat,const char* name):Symbol(ktb,name){	// creates new bitmap with given parameters
	setup(width,height,pixelFormat);
}
BitmapSymbol::BitmapSymbol(CommandBuffer& ktb,void* memory,int width,int height,KrayPixelFormat pixelFormat,const char* name):Symbol(ktb,name){
	setupSharedMemory(memory,width,height,pixelFormat);
}

BitmapSymbol::BitmapSymbol(CommandBuffer& ktb,const char* name):Symbol(ktb,name){
	ktb + KrayCmd(bitmap) + *((Symbol*)this);

	ktb.parse();
	
	Symbol::registerSymbol(KrayBitmapType);
	refreshDirectLink();
}
BitmapSymbol::BitmapSymbol(CommandBuffer& ktb,const Bitmap& bitmap,int width,int height,const char* name):Symbol(ktb,name){
	registerSymbol(width,height,bitmap.pixelFormat);
	refreshDirectLink();

	if (getDirect()){
		bitmap.copyTo(getDirect(),width,height);
	}
}
BitmapSymbol::BitmapSymbol(const char* fileName,CommandBuffer& ktb,const char* name):Symbol(ktb,name){
	ktb + KrayCmd(bitmap) + *((Symbol*)this);

	ktb + KrayEnum(bitmaps_load) + fileName;

	ktb.parse();
	
	Symbol::registerSymbol(KrayBitmapType);
	refreshDirectLink();
}

void BitmapSymbol::registerSymbol(int width,int height,KrayPixelFormat pixelFormat){
	ktb() + KrayCmd(bitmap) + *((Symbol*)this);

	switch(pixelFormat){
		case PF_FLOAT_GRAY:
			ktb() + KrayEnum(bitmaps_floatGray);
		break;

		case PF_FLOAT_RGBx:
		case PF_FLOAT_RGB:
			ktb() + KrayEnum(bitmaps_floatRgb);
		break;

		default:
		case PF_FLOAT_RGBA:
			ktb() + KrayEnum(bitmaps_floatRgba);
		break;

		case PF_CHAR_RGBA:
			ktb() + KrayEnum(bitmaps_charRgba);
		break;

		case PF_CHAR_GRAY:
			ktb() + KrayEnum(bitmaps_charGray);
		break;

		case PF_BITPLANE:
			ktb() + KrayEnum(bitmaps_bitPlane);			
		break;
	}

	kbo.pixels=0;

	(ktb() + Vector(width,height)).parse();
	
	Symbol::registerSymbol(KrayBitmapType);
}
void BitmapSymbol::refreshDirectLink(){
	if (directObjectHandle){
		ktb().getKrayInstance()->vtableBitmap()->getBitmapInfo(directObjectHandle,&kbo);
	}else{
		kbo.pixels=0;
	}
}
char BitmapSymbol::setSharedMemory(void* mem,int w,int h){
	char returnCode=0;	// ==0 fail !=0 ok

	if (directObjectHandle){
		returnCode=ktb().getKrayInstance()->vtableBitmap()->setBitmapSharedMemory(directObjectHandle,mem,w,h);

		refreshDirectLink();
	}

	return returnCode;
}
void BitmapSymbol::setPixel(int x,int y,const Vector& rgba){
	(ktb() + KrayCmd(bitmapOp) + (*this) + KrayEnum(bitmapOp_plot) + Vector(x,y) + rgba).parse();
}

//////////////////////////////////////////////////////////// VertexMap

void VertexMapSymbol::registerSymbol(int dim,size_t size,int animSamples){
	info.clear();
	
	ktb() + KrayCmd(vertexMap) + *((Symbol*)this) + dim;

	if (size){
		ktb() + KrayEnum(vertexMapTypes_fixedSize) + size + animSamples;
	}

	ktb().parse();

	Symbol::registerSymbol(KrayVertexMapType);
	refreshDirectLink();
}
void VertexMapSymbol::refreshDirectLink(){
	if (directObjectHandle){
		info.handle=(KrayVertexMap*)directObjectHandle;
		info.doublePrecision=((ktb().getKrayInstance()->getKrayLib()->getInfoInt(KRAY_INT_COMPILE_FLAGS))&KRAY_COMPILE_FLAG_VECTOR_DOUBLE)!=0;
	}else{
		info.handle=0;
	}
}
const VertexMapInfo* VertexMapSymbol::direct() const{
	if (info.ok()){
		return &info;
	}

	return 0;
}
void VertexMapSymbol::data(size_t index,const Vector& dta){
	(ktb() + (dta.count==2?KrayCmd(vertex2d):KrayCmd(vertex3d)) + (*this) + index + dta).parse();
}
void VertexMapSymbol::data(size_t index,const VarLenDouble& dta){
	(ktb() + KrayCmd(vertexData) + (*this) + index + dta).parse();
}

size_t VertexMapInfo::getSize() const{
	return handle->stop;
}
int VertexMapInfo::getDimension() const{
	return handle->dim;
}
int VertexMapInfo::getSamples() const{
	return handle->samples;
}


float* VertexMapInfo::getRowFloatPtr(size_t index) const{
	return doublePrecision?0:data<float>();
}
double* VertexMapInfo::getRowDoublePtr(size_t index) const{
	return doublePrecision?data<double>():0;
}

// those versions does automatic conversions of format
float VertexMapInfo::getValueFloat(size_t index,int column,int sample) const{
	if (doublePrecision){
		return (float)data<double>()[dataIndex(index,column,sample)];
	}else{
		return data<float>()[dataIndex(index,column,sample)];
	}
}
double VertexMapInfo::getValueDouble(size_t index,int column,int sample) const{
	if (doublePrecision){
		return data<double>()[dataIndex(index,column,sample)];
	}else{
		return data<float>()[dataIndex(index,column,sample)];
	}
}
void VertexMapInfo::setValue(size_t index,int column,float value,int sample) const{
	if (doublePrecision){
		data<double>()[dataIndex(index,column,sample)]=value;
	}else{
		data<float>()[dataIndex(index,column,sample)]=value;
	}
}
void VertexMapInfo::setValue(size_t index,int column,double value,int sample) const{
	if (doublePrecision){
		data<double>()[dataIndex(index,column,sample)]=value;
	}else{
		data<float>()[dataIndex(index,column,sample)]=(float)value;
	}
}

template <class S,class D> inline void copyFloatingPoint(const S* source,D* destination,size_t n){
	for (size_t i=0; i<n ; i++){
		destination[i]=(D)source[i];
	}
}
template <class F> void VertexMapInfo::getRowTemplate(size_t index,F* row,int sample) const{
	if (doublePrecision){
		copyFloatingPoint(data<double>()+dataIndex(index,0,sample),row,getDimension());
	}else{
		copyFloatingPoint(data<float>()+dataIndex(index,0,sample),row,getDimension());
	}
}
template <class F> void VertexMapInfo::setRowTemplate(size_t index,const F* row,int sample) const{
	if (doublePrecision){
		copyFloatingPoint(row,data<double>()+dataIndex(index,0,sample),getDimension());
	}else{
		copyFloatingPoint(row,data<float>()+dataIndex(index,0,sample),getDimension());
	}
}

void VertexMapInfo::getRow(size_t index,float* row,int sample) const{
	getRowTemplate(index,row,sample);
}
void VertexMapInfo::getRow(size_t index,double* row,int sample) const{
	getRowTemplate(index,row,sample);
}
void VertexMapInfo::setRow(size_t index,const float* row,int sample) const{
	setRowTemplate(index,row,sample);
}
void VertexMapInfo::setRow(size_t index,const double* row,int sample) const{
	setRowTemplate(index,row,sample);
}

//////////////////////////////////////////////////////////// Mesh

void MeshSymbol::registerSymbol(){
	(ktb() + KrayCmd(mesh) + *((Symbol*)this)).parse();
		
	Symbol::registerSymbol(KrayMeshType);
}
void MeshSymbol::addPoly(int polygon_tag,const VarLenDouble& polygon,int groupId,int flags){
	(ktb() + KrayCmd(meshAddPoly) + (*this) + polygon_tag + polygon + groupId + flags);
 
	ktb().parse();
}
void MeshSymbol::points(const VertexMapSymbol& v3){
 (ktb() + KrayCmd(meshOp) + (*this) + KrayEnum(meshOperations_points) + v3).parse();	
}
