#ifndef KRAY_MATRIX_H
#define KRAY_MATRIX_H

namespace Kray{

class Axes;
class AxesHpb;
class Vector;

template <int N> class MatrixTemplate{
	double matrix[N][N];

public:
	double* operator[](int v){
		return matrix[v];
	}
	const double* operator[](int v) const{
		return matrix[v];
	}

	double* getArray(){
		return &matrix[0][0];
	}
	const double* getArray() const{
		return &matrix[0][0];
	}

	template <class T> void copy(const T* a){
		for (int y=0 ; y<N ; y++){
			for (int x=0 ; x<N ; x++){
				matrix[y][x]=(double)a[x+y*N];
			}
		}
	}
	template <class T> void copyVector(int i,const T* a){
		for (int x=0 ; x<N ; x++){
			matrix[i][x]=(double)a[x];
		}
	}
	template <class T> void copyTransposed(const T* a){
		for (int y=0 ; y<N ; y++){
			for (int x=0 ; x<N ; x++){
				matrix[y][x]=(double)a[y+x*N];
			}
		}
	}

	template <class T> void copyTo(T* a) const{
		for (int y=0 ; y<N ; y++){
			for (int x=0 ; x<N ; x++){
				a[x+y*N]=(T)matrix[y][x];
			}
		}
	}
	template <class T> void copyTransposedTo(T* a) const{
		for (int y=0 ; y<N ; y++){
			for (int x=0 ; x<N ; x++){
				a[y+x*N]=(T)matrix[y][x];
			}
		}
	}

	void clear(){
		for (int y=0 ; y<N ; y++){
			for (int x=0 ; x<N ; x++){
				matrix[y][x]=0;
			}
		}
	}

	void eye(){
		clear();

		for (int i=0 ; i<N ; i++){
			matrix[i][i]=1;
		}
	}

	bool equals(const MatrixTemplate<N>& m2) const{
		for (int y=0 ; y<N ; y++){
			for (int x=0 ; x<N ; x++){
				if (matrix[y][x]!=m2.matrix[y][x]){
					return false;
				}
			}
		}
				
		return true;
	}

	void multiply(const MatrixTemplate<N>& m2,MatrixTemplate<N>& result) const{
		for (int i=0 ; i<N ; i++){
			for (int j=0 ; j<N ; j++){
				double t=0;

				for (int k=0 ; k<N ; k++){
					t+=(*this)[i][k]*m2[k][j];
				}

				result[i][j]=t;
			}
		}
	}
	void multiplyTransposed(const MatrixTemplate<N>& m2,MatrixTemplate<N>& result) const{
		for (int i=0 ; i<N ; i++){
			for (int j=0 ; j<N ; j++){
				double t=0;

				for (int k=0 ; k<N ; k++){
					t+=(*this)[i][k]*m2[j][k];
				}

				result[i][j]=t;
			}
		}
	}

	void multiply(double v){
		for (int i=0 ; i<N ; i++){
			for (int j=0 ; j<N ; j++){
				(*this)[i][j]*=v;
			}
		}
	}

	void multiply(const MatrixTemplate<N>& m2){
		MatrixTemplate<N> r;

		multiply(m2,r);

		(*this)=r;
	}
	void multiplyTransposed(const MatrixTemplate<N>& m2){
		MatrixTemplate<N> r;

		multiplyTransposed(m2,r);

		(*this)=r;
	}

	MatrixTemplate<N> operator*(const MatrixTemplate<N>& m2) const{
		MatrixTemplate<N> r;

		multiply(m2,r);

		return r;
	}
	MatrixTemplate<N>& operator*=(const MatrixTemplate<N>& m2){
		multiply(m2);

		return (*this);
	}
};

class Matrix3x3 : public MatrixTemplate<3>{
	friend class Matrix4x4;

	static void toAxes(AxesHpb* ax,Matrix3x3& t);	// t is input parameter, destroyed during computation

public:
	Matrix3x3(){
	}
	Matrix3x3(double x0,double y0,double z0, double x1,double y1,double z1, double x2,double y2,double z2);
	Matrix3x3(const AxesHpb& a){
		fromAxes(a);
	}

	void setRotation(int axisA,int axisB,double angle);

	bool isEye() const;
	bool isRightHanded() const;
	void fromAxes(const AxesHpb& a);
	void fromXYZ(double x,double y,double z);
	void fromRPY(double x,double y,double z);
	void toAxes(AxesHpb& r) const{
		Matrix3x3 t=*this;

		toAxes(&r,t);
	}
	void inverse();
	void inverse(Matrix3x3& r) const;
	void rotateVector(double v[3]) const;
	double determinant() const;

	Matrix3x3 operator*(const Matrix3x3& m2) const{
		Matrix3x3 r;

		multiply(m2,r);

		return r;
	}
};

class Matrix4x4 : public MatrixTemplate<4>{	// homegenous coordinates transformation matrix
public:
	void toMatrix3x3(Matrix3x3& t) const;

	operator AxesHpb() const;
	operator Axes() const;
	operator Vector() const;
};

};

#endif
