#ifndef KRAY_AXES_H
#define KRAY_AXES_H

#include "../base/parser.h"
#include "../math/matrix.h"
#include "vector.h"

#ifndef PI
#define PI 3.1415926535897932384626433832795
#endif

namespace Kray{

class Axes : public KrayAxesCore{		// different way of storing AxesHpb data, as 3x3 matrix
	void asAxesHpb(AxesHpb& v) const{
		asAxesHpb(*this,v);
	}
	void asMatrix(Matrix3x3& m) const{
		asMatrix(*this,m);
	}

public:
	static void asAxesHpb(const KrayAxesCore& core,AxesHpb& v);
	static void asMatrix(const KrayAxesCore& core,Matrix3x3& m);
	static void copy(KrayAxesCore& d,const KrayAxesCore& core);
	static void clear(KrayAxesCore& d){
		d.count=0;
	}
	static bool isEye(const KrayAxesCore& core);
	static void toAxesHpb(const KrayAxesCore& core,AxesHpb& v);
	static void toMatrix(const KrayAxesCore& core,Matrix3x3& m);

public:
	Axes(){
		count=0;
	}
	Axes(const Matrix3x3& m){
		(*this)=m;
	}
	Axes(const KrayAxesCore& core){
		(*this)=core;
	}
	Axes& operator=(const Matrix3x3& m){
		count=3;		// 3x3 numbers

		m.copyTo(matrix[0]);

		return (*this);
	}
	Axes& operator=(const KrayAxesCore& core){
		copy(*this,core);

		return (*this);
	}

	bool isEye() const{
		return isEye(*this);
	}
	bool isAxesHpb() const{
		return count==2;
	}
	bool isMatrix() const{
		return count==3;
	}
	bool isValid() const{
		return count>=0;
	}

	void invalidate(){
		count=-1;
	}

	void clear(){
		clear(*this);
	}

	int dim() const{
		return krayParserAxesMax;
	}

	bool isAxesHpb(AxesHpb& v) const;
	bool isMatrix(Matrix3x3& m) const;

	CoordSys toCoordSys() const;
	CoordSys toCoordSys(double x,double y,double z) const;

	void transform(KrayAxesCore& a0,char reverseTransform=0) const;

	void toAxesHpb(AxesHpb& v) const{
		return toAxesHpb(*this,v);
	}
	void toMatrix(Matrix3x3& m) const{
		return toMatrix(*this,m);
	}

	Axes& operator*=(double v){
		scale(v);

		return (*this);
	}
	Axes& scale(double v){
		return scale(v,v,v);
	}
	Axes& scale(double v0,double v1,double v2);
	void getScale(double scl[3]) const;

	KrayDtTypes dtType() const{
		return KrayDtAxesType;
	}
};

class AxesHpb : public Axes{
public:
	AxesHpb(){
		radians(0,0,0);

		setScale(1);
	}
	AxesHpb(double v0,double v1,double v2){
		radians(v0,v1,v2);

		setScale(1);
	}

	AxesHpb& radians(double v0,double v1,double v2){
		count=2;

		hpb(0)=v0;
		hpb(1)=v1;
		hpb(2)=v2;

		return *this;
	}
	AxesHpb& degrees(double v0,double v1,double v2){
		return radians(v0*PI/180,v1*PI/180,v2*PI/180);
	}
	AxesHpb& gradians(double v0,double v1,double v2){
		return radians(v0*PI/200,v1*PI/200,v2*PI/200);
	}

	AxesHpb& setScale(double v){
		return setScale(v,v,v);
	}
	AxesHpb& setScale(double v0,double v1,double v2){
		scl(0)=v0;
		scl(1)=v1;
		scl(2)=v2;

		return *this;
	}
	AxesHpb& scale(double v){	
		return scale(v,v,v);
	}
	AxesHpb& scale(double v0,double v1,double v2){
		scl(0)*=v0;
		scl(1)*=v1;
		scl(2)*=v2;

		return *this;
	}

	AxesHpb& operator*(const Vector& v){
		if (v.count==3){
			return scale(v[0],v[1],v[2]);
		}

		return *this;
	}

	void resetHpb(){
		radians(0,0,0);
	}
	void resetScl(){
		setScale(1);
	}

	bool isHpbNulls() const{
		for (int a=0 ; a<dim() ; a++){
			if (hpb(a)!=0.0){
				return false;
			}
		}
		
		return true;
	}
	bool isSclOnes() const{
		for (int a=0 ; a<dim() ; a++){
			if (scl(a)!=1.0){
				return false;
			}
		}
		
		return true;
	}
	bool isSclProportional() const{
		return scl(0)==scl(1) && scl(1)==scl(2);
	}

	bool isRightHanded() const{
		double t=scl(0);

		for (int a=1 ; a<dim() ; a++){
			t*=scl(a);
		}

		return t>0;
	}

	void eye(){
		resetHpb();
		resetScl();
	}
	bool isEye(){
		return isHpbNulls() && isSclOnes();
	}

	double* hpb(){
		return matrix[0];
	}
	const double* hpb() const{
		return matrix[0];
	}

	double& hpb(int i){
		return hpb()[i];
	}
	double hpb(int i) const{
		return hpb()[i];
	}

	double* scl(){
		return matrix[1];
	}
	const double* scl() const{
		return matrix[1];
	}

	double& scl(int i){
		return scl()[i];
	}
	double scl(int i) const{
		return scl()[i];
	}

	bool isAxes(AxesHpb& v) const{
		return true;
	}
	bool isMatrix(Matrix3x3& m) const{
		return false;
	}

	void toAxes(AxesHpb& m) const{
		m=(*this);
	}
	void toMatrix(Matrix3x3& m) const{
		m.fromAxes(*this);
	}

	AxesHpb& operator=(const Matrix3x3& m);

	Matrix3x3 operator*(const AxesHpb& v) const;
	Matrix3x3 operator*(const Matrix3x3& v) const;
	AxesHpb& operator*=(const Matrix3x3& v);
	AxesHpb& operator*=(const AxesHpb& v);
};

class AxesSymbol : public DatatypeSymbol{
	void registerSymbol();
public:
	AxesSymbol(CommandBuffer& ktb,const char* name=0):DatatypeSymbol(ktb,name){
		registerSymbol();
	}	

	void directSet(const Axes& v);
	void directGet(Axes&) const;
};

class VarLenAxes : public VarLenTemplate<AxesHpb>{
	VarLenAxes(){
	}

public:
	VarLenAxes(const AxesHpb& a0){	// creates single object view, NOTE! v0 is referenced, not copied!
		count=1;
		v=(AxesHpb*)&a0;
	}
};

template <size_t N> class VarLenAxesFixed : public VarLenFixedTemplate<VarLenAxes,N>{
};

class VarLenAxesDynamic : public VarLenDynamicTemplate<VarLenAxes>{
};

};

#endif