#ifndef ROTALLOC_H
#define ROTALLOC_H

#include "bitarray.h"

class IdAllocator{
protected:
	BitArray usedmap;
	size_t searchStartPoint;

protected:
	size_t allocInSize(){
		size_t result=usedmap.findFirst(0,lastFoundIndex(),getSize());
		
		if (result==~0){
			result=usedmap.findFirst(0,0,lastFoundIndex());
		}

		if (result!=~0){
			usedmap.set(result,1);

			setLastFoundIndex(result);
			
			return result;
		}
		
		return ~0;
	}

public:
	IdAllocator():searchStartPoint(0){
	}

	void setMinimalIdMode(char v){	// if v==1, id is always lowest possible, otherwise, rotational mode is on
		searchStartPoint=v?~0:0;
	}

	size_t getSize() const{
		return usedmap.getSize();
	}

	void resize(size_t n){
		usedmap.resize(n);
	}

	template <class T> static size_t alloc(T* ida){
		size_t result=ida->allocInSize();
		
		if (result!=~0) return result;
		
		size_t oldSize=ida->getSize();
		ida->resize(oldSize?(oldSize*2):1);
		
		return ida->allocInSize();
	}
	size_t alloc(){
		return alloc(this);
	}
	void free(size_t i){
		if (i<getSize()){
			usedmap.clear(i);
		}
	}

	void setLastFoundIndex(size_t a){
		if (searchStartPoint!=~0){
			searchStartPoint=a;
		}
	}
	size_t lastFoundIndex() const{
		if (searchStartPoint==~0) return 0;

		return searchStartPoint;
	}
};

template <class T,char zeroUnused=0> class RotationalAllocator : protected IdAllocator{
	T* data;

public:
	RotationalAllocator():data(){
	};
	~RotationalAllocator(){
		freeAll();
	}
	
	T& operator[](size_t i) const{
		return *getData(i);
	}
	T* getData(size_t index) const{
		return data+index;
	}
	char isUsed(size_t index) const{
		return usedmap.get(index);
	}

protected:
	T* allocMem(size_t size) const{
		return usedmap.allocateTable<T>(size);
	}
	void freeMem(){
		usedmap.freeTable(data);
		data=0;
	}

public:
	void setMinimalIdMode(char v){
		IdAllocator::setMinimalIdMode(v);
	}
	void setSize(size_t size){
		free();
		data=allocMem(size);
		
		usedmap.alloc(size);
	}
	size_t getSize() const{
		return IdAllocator::getSize();
	}
	void resize(size_t nSize){
		T* ndata=allocMem(nSize);

		size_t msize=BitArray::stmin(nSize,getSize());
		
		for (size_t i=0 ; i<msize ; i++){
			ndata[i]=data[i];
		}

		if (zeroUnused){
			for (size_t i=msize ; i<nSize ; i++){
				ndata[i]=0;
			}
		}
		
		freeMem();
		data=ndata;
		
		usedmap.resize(nSize);
	}
	size_t allocInSize(){
		return IdAllocator::allocInSize();
	}
	size_t alloc(){
		return IdAllocator::alloc(this);
	}
	size_t alloc(const T& value){
		size_t r=IdAllocator::alloc(this);

		(*this)[r]=value;

		return r;
	}

	void free(size_t i){
		if (i<getSize()){
			usedmap.clear(i);

			if (zeroUnused){
				(*this)[i]=0;
			}
		}
	}
	void freePointer(const T* a){
		free((size_t)(a-data));
	}
	void freeAll(){
		usedmap.freeMem();
		
		freeMem();
	}
};

template <class T> class RotationalPointerAllocator : public RotationalAllocator<T*,1>{
	typedef RotationalAllocator<T*,1> P;

public:
	RotationalPointerAllocator(){
	}
	RotationalPointerAllocator(int allocId0){
		if (allocId0){
			P::alloc(0);
		}
	}

	size_t allocNonZero(T* dta){
		if (!dta) return 0;

		return P::alloc(dta);
	}

	T* getPointer(size_t i) const{
		if (i<P::getSize()){
			return (*this)[i];
		}else{
			return 0;
		}
	}
};

#endif
